//********************************************************************
PROGRAM ShadowMapDemo_PROGRAM;

// Fenster als Popup-Fenster einrichten!
//********************************************************************
USES
   SysUtils, Classes, Math,
   Windows, Messages,
   dglOpenGL;
//********************************************************************
//   Global CONSTANTS & TYPES
//********************************************************************
TYPE
   // Basic variables
   TMatrix16d = Array[0..15] Of Double;
   TBoundingBox2D = Record Left,Top,Width,Height: LongInt; End;
   TDemoError = AnsiString;

   TWindowID = HWND;
   TDeviceContext = HDC;
   TRenderContextID = HGLRC;
   TGraphicID = GLHandle;

   TDemoWindow = Record
      Identity: TWindowID;
      BBox: TBoundingBox2D;
   End;

   TDemoRenderContext = Record
      RenderContext: TRenderContextID;
      VersionString: AnsiString;
   End;

   TDemoFBO = Record
      Container,DepthImage: TGraphicID;
      Width,Height: LongInt;
   End;

   TDemoShader = Record
      ProgramID: TGraphicID;
      VtxShaderID: TGraphicID;
      FrgShaderID: TGraphicID;
      VtxShaderText: TStringList;
      FrgShaderText: TStringList;
   End;

//********************************************************************
//   GLOBAL VARIABLES, GENERAL
//********************************************************************
VAR
   DemoWindow: TDemoWindow;
   DemoRenderContext: TDemoRenderContext;

   DoTerminate: LongBool;

   DemoLogFile: TStringList;
   DemoError: TDemoError;
   OpenGLError: TDemoError;

   DemoFBO: TDemoFBO;
   DemoShader: TDemoShader;
   ShadowMapLocation: GLint;
   Angle: LongInt;

//********************************************************************
//   GLOBAL VARIABLES, OPERATING SYSTEM
//********************************************************************
   WindowClassName: WideString;
   PixelFormatRec: TPixelFormatDescriptor;

//********************************************************************
Procedure OSInitialize;
Begin
End;
//********************************************************************
Function ReceiveEvent(AWindow: TWindowID;
                   AMessage: LongWord;
                   WParam: WPARAM; LParam: LPARAM): LongWord; StdCall;
Var
   LocalRect: TRect;
Begin
   Case AMessage Of

      WM_PAINT: Begin

         // This stops the WM_PAINT-loop
         GetUpdateRect(AWindow,LocalRect,TRUE);
         ValidateRect(AWindow,@LocalRect);
         Result:= LongInt(TRUE);
      End;

      WM_KEYDOWN: Begin  // Function keys only!
         If WParam = VK_ESCAPE
            Then DoTerminate:= TRUE;
         Result:= LongInt(TRUE);

      End;
   End; // Of Case
End;
//********************************************************************
//********************************************************************
Function gsGetErrorString: AnsiString;
Var
   InternErrorCode: GLenum; //LongWord;
Begin
   InternErrorCode:= glGetError;

   Case InternErrorCode Of

           GL_INVALID_ENUM: Result:= 'Invalid Enum';
          GL_INVALID_VALUE: Result:= 'Invalid Value';
      GL_INVALID_OPERATION: Result:= 'Invalid Operation';
          GL_OUT_OF_MEMORY: Result:= 'Out Of Memory';
               GL_NO_ERROR: Result:= 'No Error';

               Else Result:= '';

   End; // Of Case
End;
//********************************************************************
//********************************************************************
//   LOAD OPENGL LIBRARY & INIT VISUAL SETTINGS
//   Result: 0 = OK, 1 = OpenGL not loaded, 2 = PixelFormat not found
//********************************************************************
Function Load_OpenGL_And_Init_Visual_Settings: TDemoError;
Var
   WindowClass: TWndClassExW;
   OKWindowClass,OKOpenGL: LongBool;
Begin
   WindowClassName:= 'OpenGL_Demo_Window';

   With WindowClass Do
   Begin
      cbSize:= SizeOf(WindowClass);
      Style:= CS_HREDRAW Or CS_VREDRAW Or CS_OWNDC;
      lpfnWndProc:= @ReceiveEvent;
      cbClsExtra:= 0;
      cbWndExtra:= 0;
      hInstance:= System.MainInstance;
      hIcon:= LoadIcon(0, idi_Application);
      hCursor:= LoadCursor(0, idc_Arrow);
      hbrBackground:= 0;
      lpszMenuName:= NIL;
      lpszClassName:= PWideChar(WindowClassName);
      hIconSm:= LoadIcon(0, idi_Application);
   End;

   OKWindowClass:= RegisterClassExW(WindowClass) <> 0;
   If OKWindowClass
   Then Result:= ''
   Else Result:= 'WindowClass could not be created';

   // Create PixelFormatDescriptor
   If OKWindowClass
   Then Begin
      FillChar(PixelFormatRec,SizeOf(TPixelFormatDescriptor),#0);
      With PixelFormatRec Do
      Begin
         nSize       := SizeOf(TPixelFormatDescriptor);
         nVersion    := 1;
         dwFlags     := PFD_SUPPORT_OPENGL;
         dwFlags     := dwFlags OR PFD_DRAW_TO_WINDOW;
         dwFlags     := dwFlags OR PFD_DOUBLEBUFFER;
         iPixelType  := PFD_TYPE_RGBA;
         cColorBits  := 32;
         cDepthBits  := 24;
         cStencilBits:= 0;
         cAccumBits  := 0;
         cAuxBuffers := 0;
         iLayerType  := 0;
      End;


      // Load OpenGL Library
      OKOpenGL:= InitOpenGL;
      If OKOpenGL
      Then Result:= ''
      Else Result:= 'OpenGL Library could not be loaded';
   End;
End;
//********************************************************************
//********************************************************************
//   CREATE OPENGL WINDOW
//   Result: 0 = OK, 1 = OpenGL not loaded, 2 = PixelFormat not found
//********************************************************************
Function Create_OpenGL_Window: TDemoError;
Const
   ParentWindow = 0;
   MainMenu = 0;
Var
   WinStyle, ExStyle: LongWord;
Begin
   //ExStyle:= WS_EX_LEFT;
   ExStyle:= WS_EX_APPWINDOW Or WS_EX_WINDOWEDGE;
   WinStyle:= WS_OVERLAPPEDWINDOW Or
              WS_CLIPSIBLINGS Or WS_CLIPCHILDREN;

   With DemoWindow Do
   Begin
      Identity:= 0;
      With BBox Do
      Begin
         Left:=100;
         Top:= 100;
         Width:= 512;
         Height:= 512;
      End;
   End;

   With DemoWindow,BBox Do
   Identity:= CreateWindowExW
           (ExStyle,
            PWideChar(WindowClassName),
            PWideChar('OpenGL_Demo'),
            WinStyle,
            Left,Top,Width,Height,
            ParentWindow, MainMenu, System.MainInstance, NIL);

   If DemoWindow.Identity <> 0
   Then Result:= ''
   Else Result:= 'Window could not be created';
End;
//********************************************************************
//********************************************************************
//   CREATE RENDERING CONTEXT
//********************************************************************
Function Create_Rendering_Context: TDemoError;
Var
   LocalDeviceContext: TDeviceContext;
   LocalRenderContext: TRenderContextID;
   PixelFormatIndex: LongInt;
   OKPixelFormat: LongBool;
Begin
    LocalDeviceContext:= GetDC(DemoWindow.Identity);

    PixelFormatIndex:= ChoosePixelFormat
       (LocalDeviceContext, @PixelFormatRec);
    OKPixelFormat:= SetPixelFormat(LocalDeviceContext,
       PixelFormatIndex,@PixelFormatRec);

    If OKPixelFormat
    Then Begin
      LocalRenderContext:= wglCreateContext(LocalDeviceContext);
      ShowWindow(DemoWindow.Identity,SW_SHOW);
      wglMakeCurrent(LocalDeviceContext,LocalRenderContext);
    End
    Else LocalRenderContext:= 0;

    With DemoRenderContext Do If LocalRenderContext > 0
    Then Begin
       RenderContext:= LocalRenderContext;
       Result:= '';
    End
    Else Begin
       RenderContext:= 0;
       Result:= 'Rendering Context could not be activated';
    End;
End;
//********************************************************************
//********************************************************************
//   LOAD OPENGL FUNCTION POINTERS (needs a valid Rendering Context)
//   These procedures are part of dglOpenGL.pas
//********************************************************************
Function Load_OpenGL_Function_Pointers: TDemoError;
Begin
   ReadExtensions;
   ReadImplementationProperties;
   OpenGLError:= gsGetErrorString;
   If OpenGLError = 'No Error'
   Then Begin
      DemoRenderContext.VersionString:= glGetString(GL_VERSION);
      Result:= '';
   End
   Else Result:= 'Error while loading OpenGL Function Pointers: '
                     + OpenGLError;
End;
//********************************************************************
//********************************************************************
//   CREATE SHADER TEXT (Render Scene)
//********************************************************************
Function Create_Shader_Text(Var AShaderInfo: TDemoShader): TDemoError;
Begin
   With AShaderInfo.VtxShaderText Do
   Begin
      Add('varying vec3 vryNormal;');
      Add('varying vec4 vryShadowPos;');

      Add('void main(void)');
      Add('{');
      Add('gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;');
      Add('vryShadowPos = gl_TextureMatrix[7] * gl_Vertex;');
      Add('vryNormal = gl_NormalMatrix * gl_Normal;');
      Add('gl_FrontColor = gl_Color;');
      Add('}');
   End;

   With AShaderInfo.FrgShaderText Do
   Begin
      Add('uniform sampler2D uniShadowMap;');
      Add('varying vec3 vryNormal;');
      Add('varying vec4 vryShadowPos;');
      Add('vec4 ShdPosDivW;');
      Add('vec3 locNormal;');
      Add('vec3 LightDir = vec3(0.0,0.0,+1.0);');
      Add('vec4 locColor;');
      Add('float Ambient = 0.3;');

      Add('void main(void)');
      Add('{');

      Add('locNormal = normalize(vryNormal);');
      Add('float LightIntensity = max(0.0,dot(locNormal,LightDir));');

      Add('ShdPosDivW = vryShadowPos/vryShadowPos.w;');
      Add('ShdPosDivW.z += +0.0005;');

      Add('float Shadow = 1.0;');

      // Calc Shadow only if Fragment is lit
      Add('if (LightIntensity > Ambient)');
      Add('{');
         Add('ShdPosDivW = vryShadowPos/vryShadowPos.w;');
         Add('ShdPosDivW.z += 0.0005;');

         Add('float ShadowMapZ = texture2D(uniShadowMap,ShdPosDivW.st).r;');
         Add('if (ShdPosDivW.w > 0.0)');
         Add('Shadow = (ShdPosDivW.z <= ShadowMapZ) ? 1.0 : Ambient;');
      Add('}');

      Add('gl_FragColor = gl_Color;');
      Add('gl_FragColor.rgb *= LightIntensity;');
      Add('if (gl_FragColor.r > 0.75)');
      Add('gl_FragColor *= Shadow;');
      Add('}');
   End;
   Result:= '';
End;
//********************************************************************
//********************************************************************
//   CREATE SHADER PAIR (Render Scene)
//********************************************************************
Function Create_Shaders(Var AShaderInfo: TDemoShader): TDemoError;
Const
   MAX_INFO_LOG_SIZE = 1024;
Var
   InfoLogSize,TextSize: LongInt;
   LocalText,ErrorText: AnsiString;
   CompileOK,LinkOK: LongBool;
Begin

   AShaderInfo.ProgramID:= glCreateProgram;

   // VertexShader
   With AShaderInfo Do
   Begin
      VtxShaderID:= glCreateShader(GL_VERTEX_SHADER);

      LocalText:= VtxShaderText.Text;
      TextSize:= Length(LocalText);
      glShaderSource(VtxShaderID,1,@LocalText,@TextSize);

      glCompileShader(VtxShaderID);
      glGetShaderIv(VtxShaderID,GL_COMPILE_STATUS,@CompileOK);

      If CompileOK
      Then Begin
         glAttachShader(ProgramID,VtxShaderID);
         ErrorText:= '';
      End
      Else Begin
         SetLength(ErrorText,MAX_INFO_LOG_SIZE);
         glGetShaderInfoLog(VtxShaderID,MAX_INFO_LOG_SIZE,
            InfoLogSize,@ErrorText[1]);
         SetLength(ErrorText,InfoLogSize); // Cut ShaderError
      End;

      FreeAndNil(VtxShaderText);
   End;

   // FragmentShader
   If CompileOK Then With AShaderInfo Do
   Begin
      FrgShaderID:= glCreateShader(GL_FRAGMENT_SHADER);

      LocalText:= FrgShaderText.Text;
      TextSize:= Length(LocalText);
      glShaderSource(FrgShaderID,1,@LocalText,@TextSize);

      glCompileShader(FrgShaderID);
      glGetShaderIv(FrgShaderID,GL_COMPILE_STATUS,@CompileOK);

      If CompileOK
      Then Begin
         glAttachShader(ProgramID,FrgShaderID);
         ErrorText:= '';
      End
      Else Begin
         SetLength(ErrorText,MAX_INFO_LOG_SIZE);
         glGetShaderInfoLog(FrgShaderID,MAX_INFO_LOG_SIZE,
            InfoLogSize,@ErrorText[1]);
         SetLength(ErrorText,InfoLogSize); // Cut ShaderError
      End;

      FreeAndNil(FrgShaderText);
   End;

   If CompileOK Then With AShaderInfo Do
   Begin
      glLinkProgram(ProgramID);

      glGetProgramIv(ProgramID,GL_LINK_STATUS,@LinkOK);

      If LinkOK
      Then ErrorText:= ''
      Else Begin
         SetLength(ErrorText,MAX_INFO_LOG_SIZE);
         glGetProgramInfoLog(ProgramID,MAX_INFO_LOG_SIZE,
            InfoLogSize,@ErrorText[1]);
         SetLength(ErrorText,InfoLogSize); // Cut ErrorText
      End;
   End;

   If ErrorText = ''
   Then With AShaderInfo Do
      ShadowMapLocation:= glGetUniformLocation(ProgramID,'uniShadowMap');

   Result:= ErrorText;
End;
//********************************************************************
//********************************************************************
//   CREATE FRAMEBUFFER OBJECT (SHADOW MAPPING)
//********************************************************************
Function Create_Framebuffer_Object_ShadowMapping: TDemoError;
Type
   TColor4Float = Array[0..3] Of Single;
Const
   SWHITE_FLOAT: TColor4Float = (255, 255, 255, 255);
Var
   FBOState: LongWord;
Begin
   With DemoWindow.BBox Do
      glViewPort(0,0,Width,Height);

   // Set Width and Height of FBO
   With DemoFBO Do
   Begin
      Width:= DemoWindow.BBox.Width;
      Height:= DemoWindow.BBox.Height;
   End;

   // Create & bind Depth Texture
   glGenTextures(1, @DemoFBO.DepthImage);
   glBindTexture(GL_TEXTURE_2D, DemoFBO.DepthImage);

   // Set Texture filters
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);

   glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, @SWHITE_FLOAT);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

   // Set Texture properties
   glTexImage2D(GL_TEXTURE_2D, 0,
      GL_DEPTH_COMPONENT, DemoFBO.Width, DemoFBO.Height, 0,
      GL_DEPTH_COMPONENT, GL_FLOAT, NIL);

   // Unbind Depth Texture
   glBindTexture(GL_TEXTURE_2D, 0);

   //------------------------------------------------------
   // Create & bind FBO Container
   glGenFramebuffers(1,@DemoFBO.Container);
   glBindFramebuffer(GL_FRAMEBUFFER,DemoFBO.Container);

   // Attach Depth Texture to FBO as a depth component
   glFramebufferTexture2D(GL_FRAMEBUFFER,
      GL_DEPTH_ATTACHMENT,GL_TEXTURE_2D,DemoFBO.DepthImage,0);

   // Instruct openGL that we won't bind a color texture with
   // the currently bound FBO
   glDrawBuffer(GL_NONE);
   glReadBuffer(GL_NONE);

   //------------------------------------------------------
   // Check FBO State
   FBOState:= glCheckFramebufferStatus(GL_FRAMEBUFFER);

   If FBOState = GL_FRAMEBUFFER_COMPLETE
   Then Result:= ''
   Else Result:= 'Framebuffer-Object incomplete';

   // Unbind FBO Container
   glBindFramebuffer(GL_FRAMEBUFFER,0);

   Result:= '';
End;
//********************************************************************
//********************************************************************
//   DELETE FRAMEBUFFER OBJECT
//********************************************************************
Function Delete_Framebuffer_Object: TDemoError;
Begin
   With DemoFBO Do
   Begin
      glDeleteTextures(1,@DepthImage);
      glDeleteFramebuffers(1,@Container);
   End;

   Result:= '';
End;
//********************************************************************
//********************************************************************
//   SET OPENGL STATES
//********************************************************************
Function Set_OpenGL_States: TDemoError;
Begin
   glEnable(GL_CULL_FACE);
   glCullFace(GL_BACK);

   glClearColor(0.75,0.75,0.75,1);

   glEnable(GL_DEPTH_TEST);
   glDepthFunc(GL_LEQUAL);

   glShadeModel(GL_SMOOTH);

   With DemoWindow.BBox Do
      glViewPort(0,0,Width,Height);

   Result:= '';
End;
//********************************************************************
//********************************************************************
//   RENDER SCENE
//********************************************************************
Procedure Render_Scene;
Var
   AxisVal: Single;
   WallSize: Single;
Begin
   AxisVal:= Sqrt(0.5);
   WallSize:= 5;


   // Orange Wall
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix;
   glTranslatef(0,0,-3);

   glMatrixMode(GL_TEXTURE);
   glPushMatrix;
   glTranslatef(0,0,-3);


   glColor4f(1.0,0.5,0.0,1.0);
   glBegin(GL_QUADS);
      glNormal3f(+0.0,+0.0,+1.0);

      glVertex3f(-WallSize,-WallSize,0);
      glVertex3f(+WallSize,-WallSize,0);
      glVertex3f(+WallSize,+WallSize,0);
      glVertex3f(-WallSize,+WallSize,0);
   glEnd;

   glMatrixMode(GL_MODELVIEW);
   glPopMatrix;
   glMatrixMode(GL_TEXTURE);
   glPopMatrix;


   //Red Cube
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix;
   glRotatef(Angle,0,AxisVal,AxisVal);

   glMatrixMode(GL_TEXTURE);
   glPushMatrix;
   glRotatef(Angle,0,AxisVal,AxisVal);

   Inc(Angle);
   If Angle = 360
      Then Angle:= 0;

   glColor4f(1.0,0.0,0.5,1.0);
   glBegin(GL_QUADS);

      // Red Cube, Front
      glNormal3f(+0.0,+0.0,+1.0);

      glVertex3f(-1.0,-1.0,+1.0);
      glVertex3f(+1.0,-1.0,+1.0);
      glVertex3f(+1.0,+1.0,+1.0);
      glVertex3f(-1.0,+1.0,+1.0);

      // Red Cube, Back
      glNormal3f(+0.0,+0.0,-1.0);

      glVertex3f(+1.0,-1.0,-1.0);
      glVertex3f(-1.0,-1.0,-1.0);
      glVertex3f(-1.0,+1.0,-1.0);
      glVertex3f(+1.0,+1.0,-1.0);

      // Red Cube, Left
      glNormal3f(-1.0,+0.0,+0.0);

      glVertex3f(-1.0,-1.0,-1.0);
      glVertex3f(-1.0,-1.0,+1.0);
      glVertex3f(-1.0,+1.0,+1.0);
      glVertex3f(-1.0,+1.0,-1.0);

      // Red Cube, Right
      glNormal3f(+1.0,+0.0,+0.0);

      glVertex3f(+1.0,-1.0,+1.0);
      glVertex3f(+1.0,-1.0,-1.0);
      glVertex3f(+1.0,+1.0,-1.0);
      glVertex3f(+1.0,+1.0,+1.0);

      // Red Cube, Top
      glNormal3f(+0.0,+1.0,+0.0);

      glVertex3f(-1.0,+1.0,+1.0);
      glVertex3f(+1.0,+1.0,+1.0);
      glVertex3f(+1.0,+1.0,-1.0);
      glVertex3f(-1.0,+1.0,-1.0);

      // Red Cube, Bottom
      glNormal3f(+0.0,-1.0,+0.0);

      glVertex3f(-1.0,-1.0,-1.0);
      glVertex3f(+1.0,-1.0,-1.0);
      glVertex3f(+1.0,-1.0,+1.0);
      glVertex3f(-1.0,-1.0,+1.0);

   glEnd;

   glMatrixMode(GL_MODELVIEW);
   glPopMatrix;
   glMatrixMode(GL_TEXTURE);
   glPopMatrix;
End;
//********************************************************************
//********************************************************************
//   RENDER DEMO
//********************************************************************
Procedure Render_Demo;
Const
   Bias: TMatrix16d = ( 0.5, 0.0, 0.0, 0.0,
                        0.0, 0.5, 0.0, 0.0,
                        0.0, 0.0, 0.5, 0.0,
                        0.5, 0.5, 0.5, 1.0);
Var
   ViewMat,ProjMat: TMatrix16d;

   Ratio,XMax,YMax: Single;
   ZNear: Single;
   ZFar: Single;
Begin
   ZNear:= 1;
   ZFar:= 20;


   // First Step: Render from the light's position into a FBO
   glBindFramebuffer(GL_FRAMEBUFFER,DemoFBO.Container);

   // Deactivate Shader
   glUseProgram(0);

   // Clear Depth buffer & disable color rendering
   glColorMask(FALSE,FALSE,FALSE,FALSE);
   glClear(GL_DEPTH_BUFFER_BIT);

   // Cull Front Faces to avoid self-shadowing
   glCullFace(GL_FRONT);

   // Set Projection
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity;

   With DemoWindow.BBox Do
      Ratio:= Width/Height;
   YMax:= Tan(60/2*PI/180)*ZNear;
   XMax:= YMax*Ratio;
   // Origin = in the middle of the Viewport
   glFrustum(-XMax,+XMax,-YMax,+YMax,ZNear,ZFar);

   // Setup Shadow View matrix
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity;
   //gluLookAt(0,0,+4,  0,0,-1, 0,+1,0);
   glTranslatef(0,0,-4);

   // Create Shadow Matrix and load into TextureMatrix7
   glGetDoublev(GL_MODELVIEW_MATRIX,@ViewMat);
   glGetDoublev(GL_PROJECTION_MATRIX,@ProjMat);

   glActiveTexture(GL_TEXTURE7);
   glMatrixMode(GL_TEXTURE);
   glLoadIdentity;
   glMultMatrixd(@Bias[0]);
   glMultMatrixd(@ProjMat[0]);
   glMultMatrixd(@ViewMat[0]);

   //--------------
   Render_Scene;
   //--------------

   // Unbind Framebuffer
   glBindFramebuffer(GL_FRAMEBUFFER,0);




   // Second Step: Render Scene with Shadowmap Shader
   // from the camera's position
   // Clear buffers & enable color rendering
   glColorMask(TRUE,TRUE,TRUE,TRUE);
   glClear(GL_COLOR_BUFFER_BIT Or GL_DEPTH_BUFFER_BIT);

   // Switch to culling Back Faces again
   glCullFace(GL_BACK);

   // Activate Shader & Bind ShadowMap to TextureUnit7
   glUseProgram(DemoShader.ProgramID);
   glUniform1i(ShadowMapLocation,7);
   glActiveTexture(GL_TEXTURE7);
   glBindTexture(GL_TEXTURE_2D,DemoFBO.DepthImage);

   // Set View
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity;

   glRotatef(30,0,1,0);   // View Matrix: Turn to the Right 30°
   glTranslatef(6,0,-7);  // View Matrix: Go to the left(6),go back(7)

   //--------------
   Render_Scene;
   //--------------

   glBindTexture(GL_TEXTURE_2D,0);
   glUseProgram(0);

   //glxSwapBuffers(XDisplay,DemoWindow.Identity);

   SwapBuffers(GetDC(DemoWindow.Identity));

   //OpenGLError:= gsGetErrorString;
End;
//********************************************************************
//********************************************************************
//********************************************************************
//********************************************************************
//********************************************************************
//                    M A I N P R O G R A M
//********************************************************************
Var
   Msg: TMsg;
   Received: LongBool;
Begin
   //********************************************************************
   //********************************************************************
   //  BEGIN OF PROGRAM: Initialize data
   //********************************************************************
   DoTerminate:= FALSE;
   DemoLogFile:= TStringList.Create;
   Angle:= 0;

   FillChar(DemoRenderContext,SizeOf(TDemoRenderContext),#0);

   FillChar(DemoShader,SizeOf(DemoShader),#0);
   With DemoShader Do
   Begin
      VtxShaderText:= TStringList.Create;
      FrgShaderText:= TStringList.Create;
   End;

   DemoError:= Load_OpenGL_And_Init_Visual_Settings; //************
   If Length(DemoError) > 0
      Then DemoLogFile.Add(DemoError);

   If Length(DemoError) = 0
   Then Begin
      DemoError:= Create_OpenGL_Window; //*************************
      If Length(DemoError) > 0
         Then DemoLogFile.Add(DemoError);
   End;

   If Length(DemoError) = 0
   Then Begin
      DemoError:= Create_Rendering_Context; //*********************
      If Length(DemoError) > 0
         Then DemoLogFile.Add(DemoError);
   End;

   If Length(DemoError) = 0
   Then Begin
      DemoError:= Load_OpenGL_Function_Pointers; //****************
      If Length(DemoError) > 0
         Then DemoLogFile.Add(DemoError);
   End;

   If Length(DemoError) = 0
   Then Begin
      DemoError:= Create_Shader_Text(DemoShader); //***************
      If Length(DemoError) > 0
         Then DemoLogFile.Add(DemoError);
   End;

   If Length(DemoError) = 0
   Then Begin
      DemoError:= Create_Shaders(DemoShader); //*******************
      If Length(DemoError) > 0
         Then DemoLogFile.Add(DemoError);
   End;

   If Length(DemoError) = 0
   Then Begin
      DemoError:= Create_Framebuffer_Object_ShadowMapping; //******
      If Length(DemoError) > 0
         Then DemoLogFile.Add(DemoError);
   End;

   If Length(DemoError) = 0
   Then Begin
      DemoError:= Set_OpenGL_States; //****************************
      If Length(DemoError) > 0
         Then DemoLogFile.Add(DemoError);
   End;

   If Length(DemoError) = 0
      Then OpenGLError:= gsGetErrorString;

   //**************************************************
   // MAINLOOP: RENDER
   //**************************************************
   OSInitialize;

   If Length(DemoError) = 0
   Then While Not DoTerminate Do
   Begin

      Received:= PeekMessageW(Msg, 0, 0, 0, PM_REMOVE);

      If Received
      Then Begin
         TranslateMessage(Msg);
         DispatchMessageW(Msg);
      End;

      Render_Demo;

      Sleep(40);
   End;

//********************************************************************
//********************************************************************
//  END OF PROGRAM: Release data
//********************************************************************
   // Release Shader
   With DemoShader Do If ProgramID > 0
      Then glDeleteProgram(ProgramID);

   // Release FBO
   If DemoFBO.Container > 0
      Then DemoError:= Delete_Framebuffer_Object;

   // Release RenderContext
   With DemoRenderContext Do If RenderContext > 0
   Then Begin
      wglMakeCurrent(GetDC(DemoWindow.Identity),RenderContext);
      wglDeleteContext(RenderContext);
   End;

   // Release Window
   With DemoWindow Do If Identity > 0
      Then DestroyWindow(Identity);

   UnregisterClassW(PWideChar(WindowClassName),System.MainInstance);

   // Save and Release logfile
   With DemoLogFile Do
      If Count > 0
         Then SaveToFile('DemoLogFile.txt');

   DemoLogFile.Free;
End.
